(* Creme Caramel Compiler (Triple C) - Michele Caceffo - 2008 *)

open Exceptions;;
open Syntaxtree;;
open Intermediate;;
open Declaration;;

(** Here are all the functions to generate the intermediate code for the arithmetical and boolean expressions *)

(** Creates the instructions for the parameters before a call
    @param plist The parameter list
    @param dect The declarations table
    @param proct The procedures table
    @param ic The instance of the intermediateCode instance
*)
let rec genParam (plist: aexp list) (dect: dec_table) (proct: proc_table) (ic: intermediateCode): unit = match plist with
        []      ->  ()
      | a::ae   ->  genParam ae dect proct ic; ic#addInstr (PARAM, (genArithmetical a dect proct ic), Null, Null)

(** Creates the instruction for the arithmetical expressions 
    @param a The arithmetical expression to analize
    @param dect The declarations table
    @param proct The procedures table
    @param ic The intermediateCode class instance
    @return An element, which can be a Value, an Offset or a temporary register
*)
and genArithmetical (a:aexp) (dect: dec_table) (proct: proc_table) (ic: intermediateCode): element = match a with
    N(n)              ->  Val(I(n))
  | R(n)              ->  Val(F(n))
  | Var(i)            ->  (
                             if(isVarDeclared i dect) then (
                                let (_,_,stackoff) = getVarDeclaration i dect in stackoff
                             ) else raise (NON_DECLARED_VARIABLE "ERROR: Undeclared variable\n")
                            )
    | Vec(i,x)          ->  (
                             if(isVarDeclared i dect) then (
                                let xRes = genArithmetical x dect proct ic and
                                v = ic#getReg and
                                (_, _, stackoff) = getVecDeclaration i 0 (-1) dect in (
                                  let t = (match stackoff with
                                              Off(_, t, _) -> t
                                            | _ -> raise (CONFIRM_RULE "Cannot find something different from Off, as we already know we have an existing variable\n")
                                           )
                                  in
                                    if (t = DInt) then
                                      (ic#addInstr (AGET, stackoff, xRes, (Reg(v,DInt)));
                                      (Reg(v,DInt)))
                                    else
                                      (ic#addInstr (AGET, stackoff, xRes, (Reg(v,DFloat)));
                                      (Reg(v,DFloat)))
                                )
                             ) else raise (NON_DECLARED_VARIABLE "ERROR: Undeclared variable\n")
                            )
    | Mat(i,x,y)        ->  (
                             if(isVarDeclared i dect) then (
                                let xRes = genArithmetical x dect proct ic and yRes = genArithmetical y dect proct ic and
                                r1,r2,r3 = (ic#getReg),(ic#getReg),(ic#getReg) and
                                (_, dim, stackoff) = getVecDeclaration i 0 0 dect in (
                                  let t = (match stackoff with
                                              Off(_, t, _) -> t
                                            | _ -> raise (CONFIRM_RULE "Cannot find something different from Off, as we already know we have an existing variable\n")
                                           )
                                  in
                                    (
                                    ic#addInstr (MUL, (Val(I(dim))), xRes, (Reg(r1,DInt)));
                                    ic#addInstr (ADD, (Reg(r1,DInt)), yRes, (Reg(r2,DInt)));
                                    if (t = DInt) then
                                      (ic#addInstr (AGET, stackoff, (Reg(r2,DInt)), (Reg(r3,DInt)));
                                      (Reg(r3,DInt)))
                                    else
                                      (ic#addInstr (AGET, stackoff, (Reg(r2,DInt)), (Reg(r3,DFloat)));
                                      (Reg(r3,DFloat)))
                                    )
                                )
                             ) else raise (NON_DECLARED_VARIABLE "ERROR: Undeclared variable\n")
                            )
  | Sum(ex1,ex2)      ->  (
                             let reg = (ic#getReg) in (
                                let op1 = (genArithmetical ex1 dect proct ic)
                                and op2 = (genArithmetical ex2 dect proct ic)
                                in
                                let t = (match op1 with
                                            Off(_, t, _) -> t
                                          | Reg(_, t) -> t
                                          | Val(v) -> ( match v with
                                                           I(_) -> DInt
                                                         | F(_) -> DFloat
                                                      )
                                          | _ -> raise (CONFIRM_RULE "We don't have any other operations")
                                         )
                                in
                                if t = DInt then (* We check the type for one of the two operands, as the semantic correctness has already been assessed *)
                                  (ic#addInstr(ADD, op1, op2,(Reg(reg,DInt)));
                                  (Reg(reg,DInt)))
                                else
                                  (ic#addInstr(ADD, op1, op2,(Reg(reg,DFloat)));
                                  (Reg(reg,DFloat)))
                             )
                            )
    | Sub(ex1,ex2)      ->  (
                             let reg = (ic#getReg) in (
                                let op1 = (genArithmetical ex1 dect proct ic)
                                and op2 = (genArithmetical ex2 dect proct ic)
                                in
                                let t = (match op1 with
                                            Off(_, t, _) -> t
                                          | Reg(_, t) -> t
                                          | Val(v) -> ( match v with
                                                           I(_) -> DInt
                                                         | F(_) -> DFloat
                                                      )
                                          | _ -> raise (CONFIRM_RULE "We don't have any other operations")
                                         )
                                in
                                if t = DInt then (* We check the type for one of the two operands, as the semantic correctness has already been assessed *)
                                  (ic#addInstr(SUB, op1, op2,(Reg(reg,DInt)));
                                  (Reg(reg,DInt)))
                                else
                                  (ic#addInstr(SUB, op1, op2,(Reg(reg,DFloat)));
                                  (Reg(reg,DFloat)))
                             )
                            )
  | Mul(ex1,ex2)      ->  (
                             let reg = (ic#getReg) in (
                                let op1 = (genArithmetical ex1 dect proct ic)
                                and op2 = (genArithmetical ex2 dect proct ic)
                                in
                                let t = (match op1 with
                                            Off(_, t, _) -> t
                                          | Reg(_, t) -> t
                                          | Val(v) -> ( match v with
                                                           I(_) -> DInt
                                                         | F(_) -> DFloat
                                                      )
                                          | _ -> raise (CONFIRM_RULE "We don't have any other operations")
                                         )
                                in
                                if t = DInt then 
				(* We check the type for one of the two operands, as the semantic correctness has already been assessed *)
                                  (ic#addInstr(MUL, op1, op2,(Reg(reg,DInt)));
                                  (Reg(reg,DInt)))
                                else
                                  (ic#addInstr(MUL, op1, op2,(Reg(reg,DFloat)));
                                  (Reg(reg,DFloat)))
                             )
                            )
  | Div(ex1,ex2)      ->  (
                             let reg = (ic#getReg) in (
                                let op1 = (genArithmetical ex1 dect proct ic)
                                and op2 = (genArithmetical ex2 dect proct ic)
                                in
                                let t = (match op1 with
                                            Off(_, t, _) -> t
                                          | Reg(_, t) -> t
                                          | Val(v) -> ( match v with
                                                           I(_) -> DInt
                                                         | F(_) -> DFloat
                                                      )
                                          | _ -> raise (CONFIRM_RULE "We don't have any other operations")
                                         )
                                in
                                if t = DInt then (* We check the type for one of the two operands, as the semantic correctness has already been assessed *)
                                  (ic#addInstr(DIV, op1, op2,(Reg(reg,DInt)));
                                  (Reg(reg,DInt)))
                                else
                                  (ic#addInstr(DIV, op1, op2,(Reg(reg,DFloat)));
                                  (Reg(reg,DFloat)))
                             )
                            )
    | FCall(Ide(i),plist)   ->  (
                                 let pten = (Hashtbl.find proct (Ide(i))) in match pten with
                                     Building(t, _)     ->  (
                                                                 let reg = (ic#getReg) in (
                                                                 (genParam plist dect proct ic);
                                                                 match t with 
                                                                    SRet(Int)   ->  ic#addInstr (CALL, Subr(i), Null, Reg(reg,DInt));
                                                                                    Reg(reg,DInt)
                                                                  | SRet(Float) ->  ic#addInstr (CALL, Subr(i), Null, Reg(reg,DFloat));
                                                                                    Reg(reg,DFloat)
                                                                  | Void        ->  raise (SEMANTIC_ERROR "Void function")
                                                                 )
                                                                )
                                   | Subroutine(t,_,_,_)    ->  (
                                                                 let reg = (ic#getReg) in (
                                                                 (genParam plist dect proct ic);
                                                                 match t with 
                                                                    SRet(Int)   ->  ic#addInstr (CALL, Subr(i), Null, Reg(reg,DInt));
                                                                                    Reg(reg,DInt)
                                                                  | SRet(Float) ->  ic#addInstr (CALL, Subr(i), Null, Reg(reg,DFloat));
                                                                                    Reg(reg,DFloat)
                                                                  | Void        ->  raise (SEMANTIC_ERROR "Void function")
                                                                 )
                                                                )
                                )
(** Creates the instruction for the boolean expressions 
    @param a The boolean expression to analize
    @param dect The declarations table
    @param proct The procedures table
    @param ic The intermediateCode class instance
    @return An element, which can be a Value or a temporary register
*)
let rec genBoolean (a: bexp) (dect: dec_table) (proct: proc_table) (ic: intermediateCode): element = match a with
    B(b)              ->  if b then (Val(I(1))) else (Val(I(0)))
  | Equ(ex1,ex2)      ->  (
                             let reg = (ic#getReg) and 
                             (t1,t2) = (genArithmetical ex1 dect proct ic),(genArithmetical ex2 dect proct ic) in (
                                 ic#addInstr(CEQU, t1, t2, (Reg(reg,DInt)));
                                 (Reg(reg,DInt))
                             )
                            )
  | LE(ex1,ex2)       ->  (
                             let reg = (ic#getReg) and 
                             (t1,t2) = (genArithmetical ex1 dect proct ic),(genArithmetical ex2 dect proct ic) in (
                                 ic#addInstr(CGE, t2, t1, (Reg(reg,DInt)));
                                 (Reg(reg,DInt))
                             )
                            )
  | LT(ex1,ex2)       ->  (
                             let reg = (ic#getReg) and 
                             (t1,t2) = (genArithmetical ex1 dect proct ic),(genArithmetical ex2 dect proct ic) in (
                                 ic#addInstr(CG, t2, t1, (Reg(reg,DInt)));
                                 (Reg(reg,DInt))
                             )
                            )
  | Not(b)            ->  (
                             let reg = (ic#getReg) in (
                                 ic#addInstr(NOT, (genBoolean b dect proct ic), Null, (Reg(reg,DInt)));
                                 (Reg(reg,DInt))
                             )
                            )
  | And(bex1,bex2)    ->  (
                             let reg = (ic#getReg) in (
                                 ic#addInstr(AND, (genBoolean bex1 dect proct ic), (genBoolean bex2 dect proct ic), (Reg(reg,DInt)));
                                 (Reg(reg,DInt))
                             )
                            )
  | Or(bex1,bex2)     ->  (
                             let reg = (ic#getReg) in (
                                 ic#addInstr(OR, (genBoolean bex1 dect proct ic), (genBoolean bex2 dect proct ic), (Reg(reg,DInt)));
                                 (Reg(reg,DInt))
                             )
                            )
